require 'amrita2/core'

module Amrita2
  module Extention
    module CheckTrue
      def code_for_leaf_element(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "val=nil", "attrs={}") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            cg.if_("val == true") do
              cg.put_constant(element.to_s)
              cg.else_ do
                code_for_element_and_scalar(element, dyn_spec, cg)
              end
            end
          end
        end
        cg.code("module_function :#{sym}")
      end
    end
    module UseOriginalElement
      include Amrita2::Runtime
      def code_for_leaf_element(element, dyn_spec, cg, &block)
        sym = dyn_spec.sym
        cg.define_method(sym, "val=nil", "attrs={}") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            cg.if_("block_given?") do
              gen_element_with_attr(element, dyn_spec, cg)
              cg.code("e.text = #{element.text.inspect}")
              cg.code("get_mainstream << yield(e).to_s")
              cg.else_ do
                code_for_element_and_scalar(element, dyn_spec, cg)
              end
            end
          end
        end
        cg.code("module_function :#{sym}")
      end

      def code_for_element_with_child(element, dyn_spec, cg)
        cg.code("yield(#{dyn_spec.module_name})")
        cg.if_("get_substream(:#{dyn_spec.output_stream}).size == 0") do
          cg.generate_static_element_code(element, dyn_spec.option) do 
            element.each_child do |e|
              e.amrita_compile_main_code(cg)
            end
          end
          cg.else_ do
            cg.code("output_substream(:#{dyn_spec.name_for(element)})")
            element.each_child do |e|
              e.amrita_compile_main_code(cg)
            end
            cg.put_constant(end_tag(element))
          end
        end
      end

      def code_for_not_leaf_element(element, dyn_spec, cg, &block)
        if dyn_spec.matched_many? and dyn_spec.first_match?(element)
          make_selection_code(element, dyn_spec, cg, &block)
        end
        sym = dyn_spec.name_for(element)
        cg.define_method(sym, "attrs={}", "&block") do
          cg.new_context("get_substream(:#{sym})") do
            code_for_element_with_child(element, dyn_spec, cg, &block)
          end
        end
        cg.code("module_function :#{sym}")

        if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
          cg.define_module(dyn_spec.module_name) do
            cg.code("Methods = #{dyn_spec.get_methods.inspect}")
            cg.code("extend Amrita2::Core::Runtime")
            cg.define_method(sym, "val=nil", "attrs={}") do
              cg.new_context("get_substream(:#{sym})") do
                cg.code("e = yield(new_element(#{element.name.inspect}))") 
                cg.code("get_mainstream << start_tag(e)")
              end
              yield
            end
            cg.code("module_function :#{sym}")
          end
        end
      end
    end

    module ExpandByMember
      def code_for_not_leaf_element(element, dyn_spec, cg, &block)
        if dyn_spec.matched_many? and dyn_spec.first_match?(element)
          make_selection_code(element, dyn_spec, cg, &block)
        end
        sym = dyn_spec.name_for(element)
        if ! dyn_spec.matched_many? or dyn_spec.first_match?(element)
          cg.define_method(sym, "val=nil", "attrs={}", "&block") do
            cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
              code_for_element_with_child(element, dyn_spec, cg, &block)
            end
          end
          cg.code("module_function :#{sym}")
          cg.define_module(dyn_spec.module_name) do
            cg.code("Methods = #{dyn_spec.get_methods.inspect}")
            cg.code("extend Amrita2::Runtime")
            yield
          end
        end
      end

      def code_for_element_with_child(element, dyn_spec, cg)
        cg.generate_static_element_code(element, dyn_spec.option) do 
          cg.if_("block_given?") do
            cg.code("yield(#{dyn_spec.module_name})")
            cg.else_ do
              dyn_spec.children.each do |c|
                cg.code("#{dyn_spec.module_name}::#{c.sym}(val.#{c.sym})")
              end
            end
          end
          element.each_child do |e|
            e.amrita_compile_main_code(cg)
          end
        end
      end
    end

    module UseArgs
      def code_for_leaf_element(element, dyn_spec, cg, &block)
        cg.code('include Amrita2::Runtime')
        sym = dyn_spec.sym
        cg.define_method(sym, "*args", "&block") do
          cg.new_context("get_substream(:#{dyn_spec.output_stream})") do
            complie_for_args(element, cg, dyn_spec.option[:delete_id])
          end
        end
        cg.code("module_function :#{sym}")
      end

      private

      def complie_for_args(element, cg, delete_id)
        replace_args(element)
        cg.code("get_mainstream << \"#{element.to_s}\"")
      end

      def replace_args(element)
        element.delete_attribute('id')
        element.attributes.each do|k,v|
          replace_attr_args(v)
        end
        element.each_child do |c|
          case c
          when REXML::Text
            c.value = replace_text_args(c.value) if c.value
          when REXML::Element
            replace_args(c)
          end
        end
      end

      def replace_attr_args(s)
        s.gsub!(/\$(\d)/) do |ss|
          case $1
          when "0"
            '#{block.call.amrita_sanitize_as_attribute if block}'
          else
            '#{args[' + ($1.to_i-1).to_s + '].amrita_sanitize_as_attribute}'
          end
        end
      end

      def replace_text_args(s)
        s.gsub(/\$(\d)/) do |ss|
          case $1
          when "0"
            '#{block.call.amrita_sanitize if block}'
          else
            '#{args[' + ($1.to_i-1).to_s + '].amrita_sanitize}'
          end
        end
      end

      def complie_for_args_a(element, cg, delete_id)
        a = {}
        element.attributes.each do|k,v|
          a[k] = v
        end
        a.delete("id") if delete_id
        cg.code("a = #{a.inspect}")
        cg.code("a.each do |k, v|")
        cg.level_up do
          cg.code("args.each_with_index do |arg, i|")
          cg.level_up do
            cg.code("v.gsub!(\"$#\{i}\", arg)")
          end
          cg.code("end")
        end
        cg.code("end")
        vname = "e#{vcnt}"
        vcnt += 1
        cg.code("#{vname} = new_element(#{element.name.inspect}, a)")
        cg.code("start_tag(#{vname}, get_mainstream)")
        element.each_child do |c|
          case c
          when REXML::Text
            cg.code("t = #{c.to_s.inspect}")
            cg.code("args.each_with_index do |arg, i|")
            cg.level_up do
              cg.code("t.gsub!(\"$#\{i}\", arg)")
            end
            cg.code("end")
            cg.code("get_mainstream << t.amrita_sanitize")
          when REXML::Element
            complie_for_args(c, cg, delete_id, vcnt)
          end
        end
        cg.code("end_tag(#{vname}, get_mainstream)")
      end
    end

    module DynamicElementSpecExtention
      def setup_extention
        unless @option[:generator] 
          @generator.extend UseArgs if @option[:use_args]
          @generator.extend CheckTrue if @option[:check_true]
          @generator.extend UseOriginalElement if @option[:use_original_element]
          @generator.extend ExpandByMember if @option[:expand_by_member]
        end
      end
    end
  end

  module Core
    class DynamicElementSpec
      include Extention::DynamicElementSpecExtention
    end
  end
end

